@@ -1,8 +1,6 @@
-/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/fbdev/fbdev.c,v 1.44 2003/09/24 02:43:21 dawes Exp $ */
-
 /*
  * Authors:  Alan Hourihane, <alanh@fairlite.demon.co.uk>
- *	     Michel Dänzer, <michdaen@iiic.ethz.ch>
+ *	     Michel Dänzer, <michel@tungstengraphics.com>
  */
 
 #ifdef HAVE_CONFIG_H
@@ -36,24 +34,29 @@
 
 #include "xf86xv.h"
 
-#define DEBUG 0
-
-#if DEBUG
-# define TRACE_ENTER(str)       ErrorF("fbdev: " str " %d\n",pScrn->scrnIndex)
-# define TRACE_EXIT(str)        ErrorF("fbdev: " str " done\n")
-# define TRACE(str)             ErrorF("fbdev trace: " str "\n")
-#else
-# define TRACE_ENTER(str)
-# define TRACE_EXIT(str)
-# define TRACE(str)
+#ifdef XSERVER_LIBPCIACCESS
+#include <pciaccess.h>
 #endif
 
+static Bool debug = 0;
+
+#define TRACE_ENTER(str) \
+    do { if (debug) ErrorF("fbdev: " str " %d\n",pScrn->scrnIndex); } while (0)
+#define TRACE_EXIT(str) \
+    do { if (debug) ErrorF("fbdev: " str " done\n"); } while (0)
+#define TRACE(str) \
+    do { if (debug) ErrorF("fbdev trace: " str "\n"); } while (0)
+
 /* -------------------------------------------------------------------- */
 /* prototypes                                                           */
 
 static const OptionInfoRec * FBDevAvailableOptions(int chipid, int busid);
 static void	FBDevIdentify(int flags);
 static Bool	FBDevProbe(DriverPtr drv, int flags);
+#ifdef XSERVER_LIBPCIACCESS
+static Bool	FBDevPciProbe(DriverPtr drv, int entity_num,
+     struct pci_device *dev, intptr_t match_data);
+#endif
 static Bool	FBDevPreInit(ScrnInfoPtr pScrn, int flags);
 static Bool	FBDevScreenInit(int Index, ScreenPtr pScreen, int argc,
 				char **argv);
@@ -68,11 +71,6 @@ static Bool	FBDevDriverFunc(ScrnInfoPtr pScrn, xorgDriverFuncOp op,
 
 enum { FBDEV_ROTATE_NONE=0, FBDEV_ROTATE_CW=270, FBDEV_ROTATE_UD=180, FBDEV_ROTATE_CCW=90 };
 
-/*static ShadowUpdateProc updateFuncs[] =
-  { shadowUpdatePacked, shadowUpdateRotate8_270, shadowUpdateRotate8_180, shadowUpdateRotate8_90,
-    shadowUpdatePacked, shadowUpdateRotate16_270, shadowUpdateRotate16_180, shadowUpdateRotate16_90,
-    shadowUpdatePacked, shadowUpdateRotate32_270, shadowUpdateRotate32_180, shadowUpdateRotate32_90 }; */
-
 
 /* -------------------------------------------------------------------- */
 
@@ -85,8 +83,17 @@ static int pix24bpp = 0;
 #define FBDEV_VERSION		4000
 #define FBDEV_NAME		"FBDEV"
 #define FBDEV_DRIVER_NAME	"fbdev"
-#define FBDEV_MAJOR_VERSION	0
-#define FBDEV_MINOR_VERSION	2
+
+#ifdef XSERVER_LIBPCIACCESS
+static const struct pci_id_match fbdev_device_match[] = {
+    {
+	PCI_MATCH_ANY, PCI_MATCH_ANY, PCI_MATCH_ANY, PCI_MATCH_ANY,
+	0x00030000, 0x00ffffff, 0
+    },
+
+    { 0, 0, 0 },
+};
+#endif
 
 _X_EXPORT DriverRec FBDEV = {
 	FBDEV_VERSION,
@@ -99,7 +106,12 @@ _X_EXPORT DriverRec FBDEV = {
 	FBDevAvailableOptions,
 	NULL,
 	0,
-	FBDevDriverFunc
+	FBDevDriverFunc,
+
+#ifdef XSERVER_LIBPCIACCESS
+    fbdev_device_match,
+    FBDevPciProbe
+#endif
 };
 
 /* Supported "chipsets" */
@@ -115,13 +127,15 @@ static SymTabRec FBDevChipsets[] = {
 typedef enum {
 	OPTION_SHADOW_FB,
 	OPTION_ROTATE,
-	OPTION_FBDEV
+	OPTION_FBDEV,
+	OPTION_DEBUG
 } FBDevOpts;
 
 static const OptionInfoRec FBDevOptions[] = {
 	{ OPTION_SHADOW_FB,	"ShadowFB",	OPTV_BOOLEAN,	{0},	FALSE },
 	{ OPTION_ROTATE,	"Rotate",	OPTV_STRING,	{0},	FALSE },
 	{ OPTION_FBDEV,		"fbdev",	OPTV_STRING,	{0},	FALSE },
+	{ OPTION_DEBUG,		"debug",	OPTV_BOOLEAN,	{0},	FALSE },
 	{ -1,			NULL,		OPTV_NONE,	{0},	FALSE }
 };
 
@@ -141,7 +155,6 @@ static const char *fbSymbols[] = {
 
 static const char *shadowSymbols[] = {
 	"shadowAdd",
-	"shadowAlloc",
 	"shadowInit",
 	"shadowSetup",
 	"shadowUpdatePacked",
@@ -200,7 +213,7 @@ static XF86ModuleVersionInfo FBDevVersRec =
 	MODINFOSTRING1,
 	MODINFOSTRING2,
 	XORG_VERSION_CURRENT,
-	FBDEV_MAJOR_VERSION, FBDEV_MINOR_VERSION, 0,
+	PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR, PACKAGE_VERSION_PATCHLEVEL,
 	ABI_CLASS_VIDEODRV,
 	ABI_VIDEODRV_VERSION,
 	NULL,
@@ -236,10 +249,11 @@ typedef struct {
 	unsigned char*			fbmem;
 	int				fboff;
 	int				lineLength;
-	unsigned char*			shadowmem;
 	int				rotate;
 	Bool				shadowFB;
+	void				*shadow;
 	CloseScreenProcPtr		CloseScreen;
+	CreateScreenResourcesProcPtr	CreateScreenResources;
 	void				(*PointerMoved)(int index, int x, int y);
 	EntityInfoPtr			pEnt;
 	/* DGA info */
@@ -283,6 +297,55 @@ FBDevIdentify(int flags)
 	xf86PrintChipsets(FBDEV_NAME, "driver for framebuffer", FBDevChipsets);
 }
 
+
+#ifdef XSERVER_LIBPCIACCESS
+static Bool FBDevPciProbe(DriverPtr drv, int entity_num,
+			  struct pci_device *dev, intptr_t match_data)
+{
+    ScrnInfoPtr pScrn = NULL;
+
+    if (!xf86LoadDrvSubModule(drv, "fbdevhw"))
+	return FALSE;
+	    
+    xf86LoaderReqSymLists(fbdevHWSymbols, NULL);
+
+    pScrn = xf86ConfigPciEntity(NULL, 0, entity_num, NULL, NULL,
+				NULL, NULL, NULL, NULL);
+    if (pScrn) {
+	char *device;
+	GDevPtr devSection = xf86GetDevFromEntity(pScrn->entityList[0],
+						  pScrn->entityInstanceList[0]);
+
+	device = xf86FindOptionValue(devSection->options, "fbdev");
+	if (fbdevHWProbe(NULL, device, NULL)) {
+	    pScrn->driverVersion = FBDEV_VERSION;
+	    pScrn->driverName    = FBDEV_DRIVER_NAME;
+	    pScrn->name          = FBDEV_NAME;
+	    pScrn->Probe         = FBDevProbe;
+	    pScrn->PreInit       = FBDevPreInit;
+	    pScrn->ScreenInit    = FBDevScreenInit;
+	    pScrn->SwitchMode    = fbdevHWSwitchModeWeak();
+	    pScrn->AdjustFrame   = fbdevHWAdjustFrameWeak();
+	    pScrn->EnterVT       = fbdevHWEnterVTWeak();
+	    pScrn->LeaveVT       = fbdevHWLeaveVTWeak();
+	    pScrn->ValidMode     = fbdevHWValidModeWeak();
+
+	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
+		       "claimed PCI slot %d@%d:%d:%d\n", 
+		       dev->bus, dev->domain, dev->dev, dev->func);
+	    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+		       "using %s\n", device ? device : "default device");
+	}
+	else {
+	    pScrn = NULL;
+	}
+    }
+
+    return (pScrn != NULL);
+}
+#endif
+
+
 static Bool
 FBDevProbe(DriverPtr drv, int flags)
 {
@@ -310,21 +373,27 @@ FBDevProbe(DriverPtr drv, int flags)
 	
 	for (i = 0; i < numDevSections; i++) {
 	    Bool isIsa = FALSE;
+#ifndef XSERVER_LIBPCIACCESS
 	    Bool isPci = FALSE;
+#endif
 
 	    dev = xf86FindOptionValue(devSections[i]->options,"fbdev");
 	    if (devSections[i]->busID) {
+#ifndef XSERVER_LIBPCIACCESS
 	        if (xf86ParsePciBusString(devSections[i]->busID,&bus,&device,
 					  &func)) {
 		    if (!xf86CheckPciSlot(bus,device,func))
 		        continue;
 		    isPci = TRUE;
-		} else if (xf86ParseIsaBusString(devSections[i]->busID))
+		} else
+#endif
+		if (xf86ParseIsaBusString(devSections[i]->busID))
 		    isIsa = TRUE;
 		  
 	    }
 	    if (fbdevHWProbe(NULL,dev,NULL)) {
 		pScrn = NULL;
+#ifndef XSERVER_LIBPCIACCESS
 		if (isPci) {
 		    /* XXX what about when there's no busID set? */
 		    int entity;
@@ -341,7 +410,9 @@ FBDevProbe(DriverPtr drv, int flags)
 		    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
 			       "claimed PCI slot %d:%d:%d\n",bus,device,func);
 
-		} else if (isIsa) {
+		} else
+#endif
+		if (isIsa) {
 		    int entity;
 		    
 		    entity = xf86ClaimIsaSlot(drv, 0,
@@ -423,7 +494,7 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 		return FALSE;
 	default_depth = fbdevHWGetDepth(pScrn,&fbbpp);
 	if (!xf86SetDepthBpp(pScrn, default_depth, default_depth, fbbpp,
-			     Support24bppFb | Support32bppFb | SupportConvert32to24 | PreferConvert32to24))
+			     Support24bppFb | Support32bppFb | SupportConvert32to24 | SupportConvert24to32))
 		return FALSE;
 	xf86PrintDepthBpp(pScrn);
 
@@ -476,6 +547,8 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 	/* use shadow framebuffer by default */
 	fPtr->shadowFB = xf86ReturnOptValBool(fPtr->Options, OPTION_SHADOW_FB, TRUE);
 
+	debug = xf86ReturnOptValBool(fPtr->Options, OPTION_DEBUG, FALSE);
+
 	/* rotation */
 	fPtr->rotate = FBDEV_ROTATE_NONE;
 	if ((s = xf86GetOptValString(fPtr->Options, OPTION_ROTATE)))
@@ -531,15 +604,8 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 		fbdevHWUseBuildinMode(pScrn);
 	pScrn->currentMode = pScrn->modes;
 
-	if (fPtr->shadowFB)
-		pScrn->displayWidth = pScrn->virtualX;	/* ShadowFB handles this correctly */
-	else {
-		int fbbpp;
-		/* FIXME: this doesn't work for all cases, e.g. when each scanline
-			has a padding which is independent from the depth (controlfb) */
-		fbdevHWGetDepth(pScrn,&fbbpp);
-		pScrn->displayWidth = fbdevHWGetLineLength(pScrn)/(fbbpp >> 3);
-	}
+	/* First approximation, may be refined in ScreenInit */
+	pScrn->displayWidth = pScrn->virtualX;
 
 	xf86PrintModes(pScrn);
 
@@ -616,6 +682,50 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 	return TRUE;
 }
 
+
+static Bool
+FBDevCreateScreenResources(ScreenPtr pScreen)
+{
+    PixmapPtr pPixmap;
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    FBDevPtr fPtr = FBDEVPTR(pScrn);
+    Bool ret;
+
+    pScreen->CreateScreenResources = fPtr->CreateScreenResources;
+    ret = pScreen->CreateScreenResources(pScreen);
+    pScreen->CreateScreenResources = FBDevCreateScreenResources;
+
+    if (!ret)
+	return FALSE;
+
+    pPixmap = pScreen->GetScreenPixmap(pScreen);
+
+    if (!shadowAdd(pScreen, pPixmap, fPtr->rotate ?
+		   shadowUpdateRotatePackedWeak() : shadowUpdatePackedWeak(),
+		   FBDevWindowLinear, fPtr->rotate, NULL)) {
+	return FALSE;
+    }
+
+    return TRUE;
+}
+
+static Bool
+FBDevShadowInit(ScreenPtr pScreen)
+{
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    FBDevPtr fPtr = FBDEVPTR(pScrn);
+    
+    if (!shadowSetup(pScreen)) {
+	return FALSE;
+    }
+
+    fPtr->CreateScreenResources = pScreen->CreateScreenResources;
+    pScreen->CreateScreenResources = FBDevCreateScreenResources;
+
+    return TRUE;
+}
+
+
 static Bool
 FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 {
@@ -623,7 +733,7 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 	FBDevPtr fPtr = FBDEVPTR(pScrn);
 	VisualPtr visual;
 	int init_picture = 0;
-	int ret,flags,width,height;
+	int ret, flags;
 	int type;
 
 	TRACE_ENTER("FBDevScreenInit");
@@ -680,11 +790,20 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 
 	if(fPtr->rotate==FBDEV_ROTATE_CW || fPtr->rotate==FBDEV_ROTATE_CCW)
 	{
-	  height = pScrn->virtualX;
-	  width = pScrn->displayWidth = pScrn->virtualY;
-	} else {
-	  height = pScrn->virtualY;
-	  width = pScrn->virtualX;
+	  int tmp = pScrn->virtualX;
+	  pScrn->virtualX = pScrn->displayWidth = pScrn->virtualY;
+	  pScrn->virtualY = tmp;
+	} else if (!fPtr->shadowFB) {
+		/* FIXME: this doesn't work for all cases, e.g. when each scanline
+			has a padding which is independent from the depth (controlfb) */
+		pScrn->displayWidth = fbdevHWGetLineLength(pScrn) /
+				      (pScrn->bitsPerPixel / 8);
+
+		if (pScrn->displayWidth != pScrn->virtualX) {
+			xf86DrvMsg(scrnIndex, X_INFO,
+				   "Pitch updated to %d after ModeInit\n",
+				   pScrn->displayWidth);
+		}
 	}
 
 	if(fPtr->rotate && !fPtr->PointerMoved) {
@@ -692,19 +811,17 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		pScrn->PointerMoved = FBDevPointerMoved;
 	}
 
-	/* shadowfb */
-	if (fPtr->shadowFB) {
-		if ((fPtr->shadowmem = shadowAlloc(width, height,
-						   pScrn->bitsPerPixel)) == NULL) {
-		xf86DrvMsg(scrnIndex,X_ERROR,
-			   "allocation of shadow framebuffer memory failed\n");
-		return FALSE;
-	}
+	fPtr->fbstart = fPtr->fbmem + fPtr->fboff;
 
-		fPtr->fbstart   = fPtr->shadowmem;
-	} else {
-		fPtr->shadowmem = NULL;
-		fPtr->fbstart   = fPtr->fbmem + fPtr->fboff;
+	if (fPtr->shadowFB) {
+	    fPtr->shadow = xcalloc(1, pScrn->virtualX * pScrn->virtualY *
+				   pScrn->bitsPerPixel);
+
+	    if (!fPtr->shadow) {
+		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+			   "Failed to allocate shadow framebuffer\n");
+		return FALSE;
+	    }
 	}
 
 	switch ((type = fbdevHWGetType(pScrn)))
@@ -737,8 +854,11 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		case 16:
 		case 24:
 		case 32:
-			ret = fbScreenInit(pScreen, fPtr->fbstart, width, height,
-					   pScrn->xDpi, pScrn->yDpi, pScrn->displayWidth, pScrn->bitsPerPixel);
+			ret = fbScreenInit(pScreen, fPtr->shadowFB ? fPtr->shadow
+					   : fPtr->fbstart, pScrn->virtualX,
+					   pScrn->virtualY, pScrn->xDpi,
+					   pScrn->yDpi, pScrn->displayWidth,
+					   pScrn->bitsPerPixel);
 			init_picture = 1;
 			break;
 	 	default:
@@ -776,7 +896,7 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 	default:
 		xf86DrvMsg(scrnIndex, X_ERROR,
 		           "internal error: unrecognised hardware type (%d) "
-			   "encountered in FBDevScreenInit()\n");
+			   "encountered in FBDevScreenInit()\n", type);
 		ret = FALSE;
 		break;
 	}
@@ -803,11 +923,7 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 			   "Render extension initialisation failed\n");
 
-	if (fPtr->shadowFB && 
-	    (!shadowSetup(pScreen) || !shadowAdd(pScreen, NULL,
-	      fPtr->rotate ? shadowUpdateRotatePackedWeak()
-	                   : shadowUpdatePackedWeak(),
-	      FBDevWindowLinear, fPtr->rotate, NULL)) ) {
+	if (fPtr->shadowFB && !FBDevShadowInit(pScreen)) {
 	    xf86DrvMsg(scrnIndex, X_ERROR,
 		       "shadow framebuffer initialization failed\n");
 	    return FALSE;
@@ -910,8 +1026,10 @@ FBDevCloseScreen(int scrnIndex, ScreenPtr pScreen)
 	
 	fbdevHWRestore(pScrn);
 	fbdevHWUnmapVidmem(pScrn);
-	if (fPtr->shadowmem)
-		xfree(fPtr->shadowmem);
+	if (fPtr->shadow) {
+	    xfree(fPtr->shadow);
+	    fPtr->shadow = NULL;
+	}
 	if (fPtr->pDGAMode) {
 	  xfree(fPtr->pDGAMode);
 	  fPtr->pDGAMode = NULL;
@@ -919,6 +1037,7 @@ FBDevCloseScreen(int scrnIndex, ScreenPtr pScreen)
 	}
 	pScrn->vtSema = FALSE;
 
+	pScreen->CreateScreenResources = fPtr->CreateScreenResources;
 	pScreen->CloseScreen = fPtr->CloseScreen;
 	return (*pScreen->CloseScreen)(scrnIndex, pScreen);
 }
@@ -944,7 +1063,7 @@ FBDevWindowLinear(ScreenPtr pScreen, CARD32 row, CARD32 offset, int mode,
     else
       *size = fPtr->lineLength = fbdevHWGetLineLength(pScrn);
 
-    return ((CARD8 *)fPtr->fbmem + fPtr->fboff + row * fPtr->lineLength + offset);
+    return ((CARD8 *)fPtr->fbstart + row * fPtr->lineLength + offset);
 }
 
 static void
